#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Creates a haiku from backus-naur form stored in a file.

Call with a filename as a command-line parameter

The file should contain a valid bnf (not ebnf or abnf).

Copyright (c) 2009, 2023 Tor Egil Hoftun Kvæstad
"""

__author__ = "Tor Egil Hoftun Kvæstad"
__version__ = "1.1.0 2023-04-13"

import random
import argparse
import re
import sys

from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union
from re import Match


class LineNotBNFError(Exception):
    """
    An error that occurs when a line is not valid BNF.
    """


@dataclass
class Symbol:
    """
    A Symbol used in BNF. Can be either terminating or nonterminating.

    Args:
        symbol -- the string representation of this symbol
        terminating -- true if this symbol is terminating, false if no
    """
    symbol: str
    terminating: bool = False

    def __str__(self) -> str:
        return self.symbol + "(" + str(self.terminating) + ")"


class BNF:
    """The BNF."""

    def islinebnf(self, line: str) -> bool:
        """
        Checks if a line is bnf, returns True or False accordingly.

        line -- the line to be checked
        """

        matches: bool = True

        if re.match(self.bnfpattern, line) is None:
            matches = False

        return matches

    @staticmethod
    def parserightside(line: str) -> list[list[Symbol]]:
        """
        Parses the right side of the bnf, returns a list of symbols.

        line -- the right side of a bnf
        """

        tokens: list[str] = re.split(r'\|', line)
        symbols: list[list[Symbol]] = []

        for token in tokens:
            tokensymbols: list[Symbol] = []
            result: str = ""

            for character in token:
                if character == "<":
                    if result != "":
                        tokensymbols.append(Symbol(result, True))
                        result = ""
                elif character == ">":
                    tokensymbols.append(Symbol(result, False))
                    result = ""
                else:
                    result += character

            if result != "":
                tokensymbols.append(Symbol(result, True))

            symbols.append(tokensymbols)

        return symbols

    def __parseline(self, line: str) -> None:
        """
        Parses a line of bnf

        line -- the line to parse
        """

        if not self.islinebnf(line):
            raise LineNotBNFError

        result: Optional[Match[str]] = re.match(self.bnfpattern, line)

        if result is None:
            raise LineNotBNFError

        sides: Tuple[Union[str, Any], ...] = result.groups()

        self.addkey(sides[0], BNF.parserightside(sides[1]))

    def __init__(self, data: Iterable[str]) -> None:
        """
        data -- a bnf as a list of strings
        """

        if not data:
            print("Error: The provided file is empty!")
            sys.exit(1)

        self.bnf: dict[str, list[list[Symbol]]] = {} #The internal representation of the bnf
        self.bnfpattern: str = r'^\s*<(.+)>\s*::=\s*(.+)$' #The pattern to identify bnf
        self.basekey: Optional[str] = None

        lineindex = 0
        for line in data:
            lineindex += 1

            try:
                self. __parseline(line)
            except LineNotBNFError:
                print("Error: Line " + str(lineindex) + " is not valid BNF!")
                sys.exit(1)

    def addkey(self, key: str, values: list[list[Symbol]]) -> None:
        """
        Adds a line of bnf

        key -- the symbol on the left side of the bnf
        values -- a list of symbols that can be substituted for the key
        """

        if key not in self.bnf:
            self.bnf[key] = values
        else:
            self.bnf[key].extend(values)

        if self.basekey is None:
            self.basekey = key

    def __str__(self) -> str:
        representation: str = ""

        for key, symbol_list in self.bnf.items():
            representation = (representation + "\n" +
                              key + ":")

            for symbols in symbol_list:
                for symbol in symbols:
                    representation = (representation + " " +
                                      symbol.symbol)

        return representation

    def generatehaikufromkey(self, key: Optional[str]) -> str:
        """
        A recursive function that generates a haiku using a key, then returns the haiku.

        key -- the key to use to generate the haiku
        """

        haiku: str = ""

        if key is None:
            print("Error: The file does not contain a valid BNF!")
            sys.exit(1)

        line: list[Symbol] = random.choice(self.bnf[key])

        for symbol in line:
            if symbol.terminating:
                haiku += symbol.symbol
            else:
                haiku += self.generatehaikufromkey(symbol.symbol)

        return haiku

    def generatehaiku(self) -> str:
        """
        A function that generates a haiku poem, returns the resulting haiku.
        """

        return self.generatehaikufromkey(self.basekey)


def readfile(filename: str) -> Iterable[str]:
    """
    Reads a file, returns the contents of the file read.

    filename -- the name of the file to read
    """

    contents: list[str] = []

    with open(filename, "r", encoding="utf-8") as bnf_file:
        contents = bnf_file.readlines()

    return contents


def main():
    """
    The main-method.

    argv -- the command-line arguments
    """

    parser = argparse.ArgumentParser(description=("Create haikus (and other text) "
                                                  "by using a BNF pattern from a file."))
    parser.add_argument("file", help=("A file containing a BNF pattern "
                                      "that should be used to generate text."))

    args = parser.parse_args()

    file_contents = readfile(args.file)
    bnf = BNF(file_contents)

    print(bnf.generatehaiku())


if __name__ == "__main__":
    main()
